from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.warning as warnings
from tuprolog.core import Struct
from tuprolog.solve import ExecutionContext, ResolutionException, Signature

InitializationIssue = warnings.InitializationIssue

MissingPredicate = warnings.MissingPredicate


def initialization_issue(
        goal: Struct,
        context: ExecutionContext,
        cause: ResolutionException = None
) -> InitializationIssue:
    return InitializationIssue(goal, cause, context)


def missing_predicate(
        signature: Signature,
        context: ExecutionContext,
        cause=None
) -> MissingPredicate:
    return MissingPredicate(cause, context, signature)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.warning.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog.core import Term
from tuprolog.solve import ExecutionContext, Signature, current_time_instant, MAX_TIMEOUT
from tuprolog.solve.primitive import SolveRequest
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.function as _function
from typing import List, Callable


LogicFunction = _function.LogicFunction

Compute = _function.Compute

ArithmeticUtilsKt = _function.ArithmeticUtilsKt

ComputeRequest = Compute.Request

ComputeResponse = Compute.Request


@jpype.JImplements(LogicFunction)
class AbstractLogicFunction(object):
    @jpype.JOverride
    def compute(self, request: ComputeRequest) -> ComputeResponse:
        raise NotImplementedError()


def logic_function(callable: Callable[[ComputeRequest], ComputeResponse]) -> LogicFunction:
    class CallableToLogicFunctionAdapter(AbstractLogicFunction):
        def compute(self, request: ComputeRequest) -> ComputeResponse:
            return callable(request)

    return CallableToLogicFunctionAdapter()


def compute_request(
        signature: Signature,
        arguments: List[Term],
        context: ExecutionContext,
        issuing_instant: int = current_time_instant,
        max_duration: int = MAX_TIMEOUT
) -> ComputeRequest:
    return ComputeRequest(signature, arguments, context, issuing_instant, max_duration)


def compute_response(result: Term) -> ComputeResponse:
    return ComputeResponse(result)


def eval_as_expression(term: Term, request: SolveRequest, index: int = None) -> Term:
    return ArithmeticUtilsKt.evalAsExpression(term, request, index)


def eval_as_arithmetic_expression(term: Term, request: SolveRequest, index: int = None) -> Term:
    return ArithmeticUtilsKt.evalAsArithmeticExpression(term, request, index)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.function.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.unify.exception as _exceptions
from ._ktadapt import *


NoUnifyException = _exceptions.NoUnifyException

OccurCheckException = _exceptions.OccurCheckException


logger.debug("Loaded JVM classes from it.unibo.tuprolog.unify.exception.*")



import logging

# noinspection PyUnresolvedReferences
import jpype
import jpype.imports

from .libs import CLASSPATH

logging.basicConfig(level=logging.WARN)
logger = logging.getLogger('tuprolog')

jars = [str(j.resolve()) for j in CLASSPATH.glob('*.jar')]

jpype.startJVM(classpath=jars)

# noinspection PyUnresolvedReferences
from it.unibo import tuprolog as _tuprolog

Info = _tuprolog.Info

JVM_VERSION = '.'.join(map(str, jpype.getJVMVersion()))

logger.info("Started JVM v" + JVM_VERSION + " with classpath: " + str(jars))

logger.info("Using 2P-Kt v" + str(Info.VERSION))



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
from ._ktadapt import *
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve as _solve
from tuprolog.core import Indicator, Struct, Term, Substitution, EMPTY_UNIFIER, TermFormatter
from tuprolog.solve.exception import ResolutionException
from tuprolog.jvmutils import jlist, jmap, JavaSystem
from functools import singledispatch
from typing import Iterable, Mapping, Any


ExecutionContext = _solve.ExecutionContext

ExecutionContextAware = _solve.ExecutionContextAware

MutableSolver = _solve.MutableSolver

Signature = _solve.Signature

Solution = _solve.Solution

SolutionFormatter = _solve.SolutionFormatter

SolveOptions = _solve.SolveOptions

Solver = _solve.Solver

SolverFactory = _solve.SolverFactory

Time = _solve.Time


@singledispatch
def signature(name: str, arity: int, vararg: bool = False) -> Signature:
    return Signature(name, arity, vararg)


@signature.register
def _signature_from_indicator(indicator: Indicator) -> Signature:
    return Signature.fromIndicator(indicator)


@signature.register
def _signature_from_term(term: Term) -> Signature:
    return Signature.fromSignatureTerm(term)


@singledispatch
def yes_solution(
        signature: Signature,
        arguments: Iterable[Term],
        substitution: Substitution.Unifier = EMPTY_UNIFIER
) -> Solution.Yes:
    return Solution.yes(signature, jlist(arguments), substitution)


@yes_solution.register
def _yes_solution_from_query(query: Struct, substitution: Substitution.Unifier = EMPTY_UNIFIER) -> Solution.Yes:
    return Solution.yes(query, substitution)


@singledispatch
def no_solution(signature: Signature, arguments: Iterable[Term]) -> Solution.No:
    return Solution.no(signature, jlist(arguments))


@no_solution.register
def _no_solution_from_query(query: Struct) -> Solution.No:
    return Solution.no(query)


@singledispatch
def halt_solution(signature: Signature, arguments: Iterable[Term], exception: ResolutionException) -> Solution.Halt:
    return Solution.halt(signature, jlist(arguments), exception)


@halt_solution.register
def _halt_solution_from_query(query: Struct, exception: ResolutionException) -> Solution.No:
    return Solution.halt(query, exception)


def current_time_instant() -> int:
    return JavaSystem.currentTimeMillis()


def solution_formatter(term_formatter: TermFormatter = TermFormatter.prettyExpressions()) -> SolutionFormatter:
    return SolutionFormatter.of(term_formatter)


MAX_TIMEOUT: int = SolveOptions.MAX_TIMEOUT

ALL_SOLUTIONS: int = SolveOptions.ALL_SOLUTIONS


def solve_options(
        lazy: bool = True,
        timeout: int = MAX_TIMEOUT,
        limit: int = ALL_SOLUTIONS,
        custom: Mapping[str, Any] = dict(),
        **kwargs: Any
) -> SolveOptions:
    opts = dict(kwargs)
    for key in custom:
        opts[key] = custom[key]
    temp = SolveOptions.of(lazy, timeout, limit)
    if len(opts) > 0:
        return temp.setOptions(jmap(opts))
    return temp


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.*")



from .domain import *
from .evaluation import *
from .existence import *
from .instantiation import *
from .message import *
from .permission import *
from .representation import *
from .syntax import *
from .system import *
from .type import *



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.library.exception as _exception


AlreadyLoadedLibraryException = _exception.AlreadyLoadedLibraryException

LibraryException = _exception.LibraryException

NoSuchALibraryException = _exception.NoSuchALibraryException


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.library.exception.*")



from decimal import Decimal
import unittest

from tuprolog.core import big_integer, python_integer, big_decimal, python_decimal, BIG_INTEGER_ZERO, BIG_INTEGER_TEN, \
    BIG_INTEGER_ONE, BIG_INTEGER_NEGATIVE_ONE, BIG_INTEGER_TWO, BIG_DECIMAL_PI, BIG_DECIMAL_E, BIG_DECIMAL_ONE_TENTH, \
    BIG_DECIMAL_ONE_HALF, BIG_DECIMAL_ONE, BIG_DECIMAL_ZERO, BigInteger


class TestKtMath(unittest.TestCase):

    def assertSameStringRepresentation(self, expected, actual):
        self.assertEqual(str(expected), str(actual))


class TestBigInteger(TestKtMath):

    def setUp(self):
        self.large_int = 1 << 64

    def test_python_to_java(self):
        self.assertSameStringRepresentation(0, big_integer("0"))
        self.assertSameStringRepresentation(0, big_integer(0))
        self.assertSameStringRepresentation(1, big_integer("1"))
        self.assertSameStringRepresentation(1, big_integer(1))
        self.assertSameStringRepresentation(self.large_int, big_integer(self.large_int))
        self.assertSameStringRepresentation(-1, big_integer("-1"))
        self.assertSameStringRepresentation(-1, big_integer(-1))
        self.assertSameStringRepresentation(-self.large_int, big_integer(-self.large_int))
        self.assertSameStringRepresentation(0x16, big_integer("16", 16))
        self.assertSameStringRepresentation(-0x16, big_integer("-16", 16))

    def test_java_to_python(self):
        self.assertEqual(0, python_integer(big_integer("0")))
        self.assertEqual(0, python_integer(big_integer(0)))
        self.assertEqual(1, python_integer(big_integer("1")))
        self.assertEqual(1, python_integer(big_integer(1)))
        self.assertEqual(self.large_int, python_integer(big_integer(self.large_int)))
        self.assertEqual(-1, python_integer(big_integer("-1")))
        self.assertEqual(-1, python_integer(big_integer(-1)))
        self.assertEqual(-self.large_int, python_integer(big_integer(-self.large_int)))
        self.assertEqual(0x16, python_integer(big_integer("16", 16)))
        self.assertEqual(-0x16, python_integer(big_integer("-16", 16)))

    def test_constants(self):
        self.assertEqual(big_integer(0), BIG_INTEGER_ZERO)
        self.assertEqual(big_integer(10), BIG_INTEGER_TEN)
        self.assertEqual(big_integer(1), BIG_INTEGER_ONE)
        self.assertEqual(big_integer(-1), BIG_INTEGER_NEGATIVE_ONE)
        self.assertEqual(big_integer(2), BIG_INTEGER_TWO)

    def test_integers(self):
        for i in range(-512, 513):
            self.assertEqual(BigInteger.of(i), big_integer(i))
        for t in [8, 16, 32, 64, 128]:
            for i in [1 << t - 1, 1 << t]:
                self.assertEqual(BigInteger.of(str(i)), big_integer(i))
                self.assertEqual(BigInteger.of(str(-i)), big_integer(-i))


class TestBigDecimal(TestKtMath):

    def setUp(self):
        self.e = '2.71828182845904523536028747135266249775724709369995957496696762772407663035354759457138217852' + \
                 '5166427427466391932003059921817413596629043572900334295260595630738132328627943490763233829880' + \
                 '753195251019011573834187930702154089149934884167509244761460668082264'

        self.pi = '3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253' + \
                  '421170679821480865132823066470938446095505822317253594081284811174502841027019385211055596446' + \
                  '22948954930381964428810975665933446128475648233786783165271201909145648'

        self.one_tenth = "0.01000000000000000020816681711721685132943093776702880859375"

    def test_python_to_java_zero(self):
        self.assertSameStringRepresentation(0, big_decimal("0"))
        self.assertSameStringRepresentation(0, big_decimal("0.0"))
        self.assertSameStringRepresentation(0, big_decimal("0.00"))
        self.assertSameStringRepresentation(0, big_decimal(0))
        self.assertSameStringRepresentation(0, big_decimal(0.0))
        self.assertSameStringRepresentation(0, big_decimal(0.00))
        self.assertSameStringRepresentation(0, big_decimal(Decimal(0)))
        self.assertSameStringRepresentation(0, big_decimal(Decimal(0.0)))
        self.assertSameStringRepresentation(0, big_decimal(Decimal(0.00)))

    def test_python_to_java_one(self):
        self.assertSameStringRepresentation(1, big_decimal("1"))
        self.assertSameStringRepresentation(1, big_decimal("1.0"))
        self.assertSameStringRepresentation(1, big_decimal("1.00"))
        self.assertSameStringRepresentation(1, big_decimal(1))
        self.assertSameStringRepresentation(1, big_decimal(1.0))
        self.assertSameStringRepresentation(1, big_decimal(1.00))
        self.assertSameStringRepresentation(1, big_decimal(Decimal(1)))
        self.assertSameStringRepresentation(1, big_decimal(Decimal(1.0)))
        self.assertSameStringRepresentation(1, big_decimal(Decimal(1.00)))

    def test_python_to_java_minus_one(self):
        self.assertSameStringRepresentation(-1, big_decimal("-1"))
        self.assertSameStringRepresentation(-1, big_decimal("-1.0"))
        self.assertSameStringRepresentation(-1, big_decimal("-1.00"))
        self.assertSameStringRepresentation(-1, big_decimal(-1))
        self.assertSameStringRepresentation(-1, big_decimal(-1.0))
        self.assertSameStringRepresentation(-1, big_decimal(-1.00))
        self.assertSameStringRepresentation(-1, big_decimal(Decimal(-1)))
        self.assertSameStringRepresentation(-1, big_decimal(Decimal(-1.0)))
        self.assertSameStringRepresentation(-1, big_decimal(Decimal(-1.00)))

    def test_python_to_java_one_half(self):
        self.assertSameStringRepresentation(0.5, big_decimal("0.5"))
        self.assertSameStringRepresentation(0.5, big_decimal("0.50"))
        self.assertSameStringRepresentation(0.5, big_decimal("0.500"))
        self.assertSameStringRepresentation(0.5, big_decimal(0.5))
        self.assertSameStringRepresentation(0.5, big_decimal(0.50))
        self.assertSameStringRepresentation(0.5, big_decimal(0.500))
        self.assertSameStringRepresentation(0.5, big_decimal(Decimal(0.5)))
        self.assertSameStringRepresentation(0.5, big_decimal(Decimal(0.50)))
        self.assertSameStringRepresentation(0.5, big_decimal(Decimal(0.500)))

    def test_python_to_java_one_cent(self):
        self.assertSameStringRepresentation(0.01, big_decimal("0.01"))
        self.assertSameStringRepresentation(0.01, big_decimal("0.010"))
        self.assertSameStringRepresentation(0.01, big_decimal("0.0100"))
        self.assertSameStringRepresentation(0.01, big_decimal(0.01))
        self.assertSameStringRepresentation(0.01, big_decimal(0.010))
        self.assertSameStringRepresentation(0.01, big_decimal(0.0100))
        self.assertSameStringRepresentation(self.one_tenth, big_decimal(Decimal(0.01)))
        self.assertSameStringRepresentation(self.one_tenth, big_decimal(Decimal(0.010)))
        self.assertSameStringRepresentation(self.one_tenth, big_decimal(Decimal(0.0100)))

    def test_java_to_python_zero(self):
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal("0")))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal("0.0")))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal("0.00")))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal(0)))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal(0.0)))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal(0.00)))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal(Decimal(0))))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal(Decimal(0.0))))
        self.assertSameStringRepresentation(Decimal(0), python_decimal(big_decimal(Decimal(0.00))))

    def test_java_to_python_one(self):
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal("1")))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal("1.0")))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal("1.00")))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal(1)))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal(1.0)))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal(1.00)))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal(Decimal(1))))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal(Decimal(1.0))))
        self.assertSameStringRepresentation(Decimal(1), python_decimal(big_decimal(Decimal(1.00))))

    def test_java_to_python_minus_one(self):
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal("-1")))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal("-1.0")))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal("-1.00")))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal(-1)))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal(-1.0)))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal(-1.00)))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal(Decimal(-1))))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal(Decimal(-1.0))))
        self.assertSameStringRepresentation(Decimal(-1), python_decimal(big_decimal(Decimal(-1.00))))

    def test_java_to_python_one_half(self):
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal("0.5")))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal("0.50")))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal("0.500")))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal(0.5)))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal(0.50)))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal(0.500)))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal(Decimal(0.5))))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal(Decimal(0.50))))
        self.assertSameStringRepresentation(Decimal(0.5), python_decimal(big_decimal(Decimal(0.500))))

    def test_java_to_python_one_cent(self):
        self.assertSameStringRepresentation(Decimal("0.01"), python_decimal(big_decimal("0.01")))
        self.assertSameStringRepresentation(Decimal("0.01"), python_decimal(big_decimal("0.010")))
        self.assertSameStringRepresentation(Decimal("0.01"), python_decimal(big_decimal("0.0100")))
        self.assertSameStringRepresentation(Decimal("0.01"), python_decimal(big_decimal(0.01)))
        self.assertSameStringRepresentation(Decimal("0.01"), python_decimal(big_decimal(0.010)))
        self.assertSameStringRepresentation(Decimal("0.01"), python_decimal(big_decimal(0.0100)))
        self.assertSameStringRepresentation(Decimal(self.one_tenth), python_decimal(big_decimal(Decimal(0.01))))
        self.assertSameStringRepresentation(Decimal(self.one_tenth), python_decimal(big_decimal(Decimal(0.010))))
        self.assertSameStringRepresentation(Decimal(self.one_tenth), python_decimal(big_decimal(Decimal(0.0100))))

    def test_constants(self):
        self.assertEqual(big_decimal(0), BIG_DECIMAL_ZERO)
        self.assertEqual(big_decimal(1), BIG_DECIMAL_ONE)
        self.assertEqual(big_decimal(0.5), BIG_DECIMAL_ONE_HALF)
        self.assertEqual(big_decimal("0.1"), BIG_DECIMAL_ONE_TENTH)
        self.assertEqual(big_decimal(self.e), BIG_DECIMAL_E)
        self.assertEqual(big_decimal(self.pi), BIG_DECIMAL_PI)


if __name__ == '__main__':
    unittest.main()



from tuprolog import logger
import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.unify.Unificator")
class _KtUnificator:
    def __jclass_init__(cls):
        pass

    @property
    def context(self):
        return self.getContext()


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.unify.*")



# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog import logger
from tuprolog.core import Var
from tuprolog.solve import ExecutionContext, Signature
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors


InstantiationError = errors.InstantiationError


def instantiation_error_for_argument(
        context: ExecutionContext,
        procedure: Signature,
        variable: Var,
        index: int = None
) -> InstantiationError:
    return InstantiationError.forArgument(context, procedure, variable, index)


def instantiation_error_for_goal(
        context: ExecutionContext,
        procedure: Signature,
        variable: Var
) -> InstantiationError:
    return InstantiationError.forGoal(context, procedure, variable)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.InstantiationError.*")



from tuprolog import logger
from tuprolog.solve.flags import DEFAULT_FLAG_STORE, FlagStore
from tuprolog.solve.library import libraries, Libraries
from tuprolog.solve.channel import InputChannel, OutputChannel, std_out, std_in, std_err, warn
from tuprolog.theory import theory, mutable_theory, Theory
from tuprolog.solve import Solver, SolverFactory


_PROLOG_SOLVER_FACTORY = Solver.getProlog()


def prolog_solver(
        libraries: Libraries = libraries(),
        flags: FlagStore = DEFAULT_FLAG_STORE,
        static_kb: Theory = theory(),
        dynamic_kb: Theory = mutable_theory(),
        std_in: InputChannel = std_in(),
        std_out: OutputChannel = std_out(),
        std_err: OutputChannel = std_err(),
        warning: OutputChannel = warn(),
        mutable: bool = True
) -> Solver:
    if mutable:
        return _PROLOG_SOLVER_FACTORY.mutableSolverWithDefaultBuiltins(
            libraries, flags, static_kb, dynamic_kb, std_in, std_out, std_err, warning
        )
    else:
        return _PROLOG_SOLVER_FACTORY.solverWithDefaultBuiltins(
            libraries, flags, static_kb, dynamic_kb, std_in, std_out, std_err, warning
        )


def prolog_solver_factory() -> SolverFactory:
    return _PROLOG_SOLVER_FACTORY


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.prolog.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.data as _data
from tuprolog.jvmutils import Map
from tuprolog.pyutils import dict_or_keyword_args
from tuprolog.jvmutils import jmap
from typing import Mapping, Any


CustomDataStore = _data.CustomDataStore

CustomData = Map


def custom_data(data: Mapping[str, Any] = {}, **kwargs) -> CustomData:
    return dict_or_keyword_args(data, kwargs, lambda ds: jmap(ds))


EMPTY_DATA: CustomData = custom_data()


def custom_data_store(
        persistent: CustomData = EMPTY_DATA,
        durable: CustomData = EMPTY_DATA,
        ephemeral: CustomData = EMPTY_DATA
) -> CustomDataStore:
    return CustomDataStore.empty().copy(persistent, durable, ephemeral)


EMPTY_DATA_STORE: CustomDataStore = CustomDataStore.empty()


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.data.*")



import unittest
from tuprolog.core import scope, var, Scope, variables, struct


class TestScope(unittest.TestCase):

    def test_empty_scope_creation(self):
        empty_scope = scope()
        self.assertIsNotNone(empty_scope)
        self.assertIsInstance(empty_scope, Scope)
        self.assertEqual(0, len(empty_scope.variables))

    def test_scope_creation(self):
        full_scope = scope('A', var('B'))
        self.assertIsNotNone(full_scope)
        self.assertIsInstance(full_scope, Scope)
        self.assertEqual(2, len(full_scope.variables))
        for v in full_scope.variables:
            self.assertIn(v, {'A', 'B'})

    def test_variables_reuse(self):
        A, B = variables('A', 'B')
        s = scope(A, B)
        self.assertEqual(A, s['A'])
        self.assertEqual(B, s['B'])
        fAB = s.struct('f', s.var('A'), s.var('B'))
        self.assertEqual(struct('f', A, B), fAB)


if __name__ == '__main__':
    unittest.main()



import unittest
from tuprolog.core import *
from tuprolog.core.formatters import AbstractTermFormatter


class TestFormatters(unittest.TestCase):

    def test_term_formatter(self):
        class MyFormatter(AbstractTermFormatter):
            def defaultValue(self, term):
                return str(term)
        formatter = MyFormatter()
        for term in [Atom.of('a'), Integer.of(1), Var.of("X")]:
            self.assertEqual(formatter.format(term), str(term))


if __name__ == '__main__':
    unittest.main()



from tuprolog import logger
from ._ktadapt import *
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.utils as _utils

Taggable = _utils.Taggable

logger.debug("Loaded JVM classes from it.unibo.tuprolog.utils.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyProtectedMember
from _jpype import _JObject as JObjectClass
# noinspection PyUnresolvedReferences
import java.util as _jutils
# noinspection PyUnresolvedReferences
import java.lang as _jlang
# noinspection PyUnresolvedReferences
import kotlin as _kotlin
# noinspection PyUnresolvedReferences
import kotlin.sequences as _ksequences
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.utils as _tuprolog_utils

from typing import Iterable as PyIterable
from typing import Iterator as PyIterator
from typing import Mapping, MutableMapping, Callable, Any

from .jvmioutils import *


Arrays = _jutils.Arrays

ArrayList = _jutils.ArrayList

Iterator = _jutils.Iterator

Map = _jutils.Map

NoSuchElementException = _jutils.NoSuchElementException

Iterable = _jlang.Iterable

JavaSystem = _jlang.System

Object = _jlang.Object

Pair = _kotlin.Pair

Triple = _kotlin.Triple

Sequence = _ksequences.Sequence

SequencesKt = _ksequences.SequencesKt

PyUtils = _tuprolog_utils.PyUtils


def protect_iterable(iterable: Iterable) -> Iterable:
    return PyUtils.iterable(iterable)


@jpype.JImplements("java.util.Iterator", deferred=True)
class _IteratorAdapter(object):
    def __init__(self, iterator):
        assert isinstance(iterator, PyIterator)
        self._iterator = iterator
        self._queue = None

    @jpype.JOverride
    def hasNext(self):
        if self._queue is None:
            try:
                self._queue = [next(self._iterator)]
                return True
            except StopIteration:
                return False
        elif len(self._queue) > 0:
            return True
        else:
            try:
                self._queue.append(next(self._iterator))
                return True
            except StopIteration:
                return False

    @jpype.JOverride
    def next(self):
        if self.hasNext():
            return self._queue.pop(0)
        else:
            raise NoSuchElementException()


@jpype.JImplements("java.lang.Iterable", deferred=True)
class _IterableAdapter(object):
    def __init__(self, iterable):
        assert isinstance(iterable, PyIterable)
        self._iterable = iterable

    @jpype.JOverride
    def iterator(self):
        return _IteratorAdapter(iter(self._iterable))


def kpair(items: PyIterable) -> Pair:
    if isinstance(items, Pair):
        return items
    i = iter(items)
    first = next(i)
    second = next(i)
    return Pair(first, second)


@jpype.JConversion("kotlin.Pair", instanceof=PyIterable, excludes=str)
def _kt_pair_covert(jcls, obj):
    return kpair(obj)


def ktriple(items: PyIterable) -> Triple:
    if isinstance(items, Triple):
        return items
    i = iter(items)
    first = next(i)
    second = next(i)
    third = next(i)
    return Triple(first, second, third)


@jpype.JConversion("kotlin.Triple", instanceof=PyIterable, excludes=str)
def _kt_triple_covert(jcls, obj):
    return ktriple(obj)


def jlist(iterable: PyIterable) -> Iterable:
    assert isinstance(iterable, PyIterable)
    if isinstance(iterable, list):
        return Arrays.asList(iterable)
    lst = ArrayList()
    for item in iterable:
        lst.add(item)
    return lst


def jiterable(iterable: PyIterable) -> Iterable:
    assert isinstance(iterable, PyIterable)
    return _IterableAdapter(iterable)


@jpype.JConversion("java.lang.Iterable", instanceof=PyIterable, excludes=str)
def _java_iterable_convert(jcls, obj):
    return jiterable(obj)


def jarray(type, rank: int = 1):
    return jpype.JArray(type, rank)


def jiterator(iterator: PyIterator) -> Iterator:
    assert isinstance(iterator, PyIterator)
    return _IteratorAdapter(iterator)


def jmap(dictionary: Mapping) -> Map:
    assert isinstance(dictionary, Mapping)
    return Map@dictionary


def _java_obj_repr(java_object: Object) -> str:
    return str(java_object.toString())


# replaces the default __repr__ implementation for java objects, making them use _java_obj_repr
JObjectClass.__repr__ = _java_obj_repr


@jpype.JImplementationFor("kotlin.sequences.Sequence")
class _KtSequence:
    def __jclass_init__(self):
        PyIterable.register(self)

    def __iter__(self):
        return protect_iterable(self).iterator()


def ksequence(iterable: PyIterable) -> Sequence:
    return SequencesKt.asSequence(jiterable(iterable))


@jpype.JConversion("kotlin.sequences.Sequence", instanceof=PyIterable, excludes=str)
def _kt_sequence_convert(jcls, obj):
    return ksequence(obj)


@jpype.JImplementationFor("java.util.stream.Stream")
class _JvmStream:
    def __jclass_init__(self):
        PyIterable.register(self)

    def __iter__(self):
        return self.iterator()


@jpype.JImplementationFor("java.lang.Comparable")
class _JvmComparable:
    def __jclass_init__(self):
        pass

    def __lt__(self, other):
        return self.compareTo(other) < 0

    def __gt__(self, other):
        return self.compareTo(other) > 0

    def __le__(self, other):
        return self.compareTo(other) <= 0

    def __ge__(self, other):
        return self.compareTo(other) >= 0


@jpype.JImplementationFor("java.lang.Throwable")
class _JvmThrowable:
    def __jclass_init__(self):
        pass

    @property
    def message(self):
        return self.getMessage()

    @property
    def localized_message(self):
        return self.getLocalizedMessage()

    @property
    def cause(self):
        return self.getCause()


class _KtFunction(Callable):
    def __init__(self, arity: int, function: Callable):
        self._function = function
        self._arity = arity

    def invoke(self, *args):
        assert len(args) == self._arity
        return self._function(*args)

    def __call__(self, *args):
        return self.invoke(*args)


_kt_function_classes: MutableMapping[int, Any] = dict()


def kfunction(arity: int):
    if arity not in _kt_function_classes:
        @jpype.JImplements("kotlin.jvm.functions.Function" + str(arity), deferred=True)
        class _KtFunctionN(_KtFunction):
            def __init__(self, f):
                super().__init__(arity, f)

            @jpype.JOverride
            def invoke(self, *args):
                return super().invoke(*args)

        _kt_function_classes[arity] = _KtFunctionN
    return _kt_function_classes[arity]


logger.debug("Configure JVM-specific extensions")



from tuprolog import logger
import jpype
from tuprolog.core import TermVisitor


@jpype.JImplements(TermVisitor)
class AbstractTermVisitor(object):

    @jpype.JOverride
    def defaultValue(self, term):
        raise NotImplementedError()

    @jpype.JOverride
    def visitTerm(self, term):
        return TermVisitor.visitTerm(self, term)

    @jpype.JOverride
    def visitVar(self, term):
        return TermVisitor.visitVar(self, term)

    @jpype.JOverride
    def visitConstant(self, term):
        return TermVisitor.visitConstant(self, term)

    @jpype.JOverride
    def visitStruct(self, term):
        return TermVisitor.visitStruct(self, term)

    @jpype.JOverride
    def visitCollection(self, term):
        return TermVisitor.visitCollection(self, term)

    @jpype.JOverride
    def visitAtom(self, term):
        return TermVisitor.visitAtom(self, term)

    @jpype.JOverride
    def visitTruth(self, term):
        return TermVisitor.visitTruth(self, term)

    @jpype.JOverride
    def visitNumeric(self, term):
        return TermVisitor.visitNumeric(self, term)

    @jpype.JOverride
    def visitInteger(self, term):
        return TermVisitor.visitInteger(self, term)

    @jpype.JOverride
    def visitReal(self, term):
        return TermVisitor.visitReal(self, term)

    @jpype.JOverride
    def visitBlock(self, term):
        return TermVisitor.visitBlock(self, term)

    @jpype.JOverride
    def visitEmpty(self, term):
        return TermVisitor.visitEmpty(self, term)

    @jpype.JOverride
    def visitEmptyBlock(self, term):
        return TermVisitor.visitEmptyBlock(self, term)

    @jpype.JOverride
    def visitList(self, term):
        return TermVisitor.visitList(self, term)

    @jpype.JOverride
    def visitCons(self, term):
        return TermVisitor.visitCons(self, term)

    @jpype.JOverride
    def visitEmptyList(self, term):
        return TermVisitor.visitEmptyList(self, term)

    @jpype.JOverride
    def visitTuple(self, term):
        return TermVisitor.visitTuple(self, term)

    @jpype.JOverride
    def visitIndicator(self, term):
        return TermVisitor.visitIndicator(self, term)

    @jpype.JOverride
    def visitClause(self, term):
        return TermVisitor.visitClause(self, term)

    @jpype.JOverride
    def visitRule(self, term):
        return TermVisitor.visitRule(self, term)

    @jpype.JOverride
    def visitFact(self, term):
        return TermVisitor.visitFact(self, term)

    @jpype.JOverride
    def visitDirective(self, term):
        return TermVisitor.visitDirective(self, term)


logger.debug("Loaded compatibility layer for JVM subtypes of " + str(TermVisitor.class_.getName()))



from tuprolog import logger
import jpype
from tuprolog.core import indicator as new_indicator
from tuprolog.jvmutils import jiterable, protect_iterable
from tuprolog.pyutils import iterable_or_varargs
from typing import Sized


@jpype.JImplementationFor("it.unibo.tuprolog.theory.Theory")
class _KtTheory:
    def __jclass_init__(cls):
        Sized.register(cls)

    @property
    def is_mutable(self):
        return self.isMutable()

    def to_mutable_theory(self):
        return self.toMutableTheory()

    def to_immutable_theory(self):
        return self.toImmutableTheory()

    @jpype.JOverride
    def getClauses(self):
        return protect_iterable(self.getClauses_())

    @property
    def clauses(self):
        return self.getClauses()

    @jpype.JOverride
    def getRules(self):
        return protect_iterable(self.getRules_())

    @property
    def rules(self):
        return self.getRules()

    @jpype.JOverride
    def getDirectives(self):
        return protect_iterable(self.getDirectives_())

    @property
    def directives(self):
        return self.getDirectives()

    def __len__(self):
        return self.getSize()

    def __add__(self, other):
        return self.plus(other)

    def __contains__(self, item):
        return self.contains(item)

    def __getitem__(self, item):
        return self.get(item)

    def _assert(self, method, clause, *clauses):
        if len(clauses) == 0:
            return method(clause)
        return iterable_or_varargs((clause,) + clauses, lambda cs: method(jiterable(cs)))

    def assert_a(self, clause, *clauses):
        self._assert(self.assertA, clause, *clauses)

    def assert_z(self, clause, *clauses):
        self._assert(self.assertZ, clause, *clauses)

    @jpype.JOverride
    def retract(self, clause, *clauses):
        if len(clauses) == 0:
            return self.retract_(clause)
        return iterable_or_varargs((clause,) + clauses, lambda cs: self.retract_(jiterable(cs)))

    def retract_all(self, clause):
        return self.retractAll(clause)

    @jpype.JOverride
    def abolish(self, name, arity=None, indicator=None):
        if name is not None:
            if arity is not None:
                return self.abolish_(new_indicator(name, arity))
            else:
                return self.abolish_(name)
        elif indicator is not None:
            return self.abolish_(indicator)
        raise ValueError("You should provide at least either a name-arity couple or an indicator")

    @jpype.JOverride
    def equals(self, other, use_var_complete_name=True):
        return self.equals_(other, use_var_complete_name)

    def __eq__(self, other):
        return self.equals(other, use_var_complete_name=True)

    def to_string(self, as_prolog_text=False):
        return self.toString(as_prolog_text)


@jpype.JImplementationFor("it.unibo.tuprolog.theory.RetractResult")
class _KtRetractResult:
    def __jclass_init__(cls):
        pass

    @property
    def is_success(self):
        return self.isSuccess()

    @property
    def is_failure(self):
        return self.isFailure()

    @property
    def theory(self):
        return self.getTheory()

    @property
    def clauses(self):
        return self.getClauses()

    @property
    def first_clause(self):
        return self.getFirstClause()


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.theory.*")
